@@ -2,6 +2,8 @@
 
 use super::{is_older_version, Key, KeyData};
 use std;
+#[cfg(feature = "unstable")]
+use std::collections::TryReserveError;
 use std::hint::unreachable_unchecked;
 use std::iter::{Enumerate, Extend, FromIterator, FusedIterator};
 use std::marker::PhantomData;
@@ -10,7 +12,7 @@ use std::ops::{Index, IndexMut};
 // We could use unions to remove the memory overhead of Option here as well, but
 // until non-Copy elements inside unions stabilize it's better to give users at
 // least some place to store non-Copy elements.
-#[derive(Debug)]
+#[derive(Debug, Clone)]
 struct Slot<T> {
     version: u32,
     value: Option<T>,
@@ -81,7 +83,7 @@ impl<T> Slot<T> {
 /// ammo[alice] = 0;
 /// ```
 
-#[derive(Debug)]
+#[derive(Debug, Clone)]
 pub struct SecondaryMap<K: Key, V> {
     slots: Vec<Slot<V>>,
     num_elems: usize,
@@ -201,6 +203,34 @@ impl<K: Key, V> SecondaryMap<K, V> {
         }
     }
 
+    /// Tries to set the capacity of the `SecondaryMap` to `new_capacity`, if it
+    /// is bigger than the current capacity.
+    ///
+    /// It is recommended to set the capacity of a `SecondaryMap` to the
+    /// capacity of its corresponding slot map before inserting many new
+    /// elements to prevent frequent reallocations. The collection may reserve
+    /// more space than requested.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// # use slotmap::*;
+    /// let mut sec: SecondaryMap<DefaultKey, i32> = SecondaryMap::with_capacity(10);
+    /// assert!(sec.capacity() >= 10);
+    /// sec.try_set_capacity(1000).unwrap();
+    /// assert!(sec.capacity() >= 1000);
+    /// ```
+    #[cfg(feature = "unstable")]
+    pub fn try_set_capacity(&mut self, new_capacity: usize) -> Result<(), TryReserveError> {
+        let new_capacity = new_capacity + 1; // Sentinel.
+        if new_capacity > self.slots.capacity() {
+            let needed = new_capacity - self.slots.len();
+            self.slots.try_reserve(needed)
+        } else {
+            Ok(())
+        }
+    }
+
     /// Returns `true` if the secondary map contains `key`.
     ///
     /// # Examples
@@ -242,12 +272,11 @@ impl<K: Key, V> SecondaryMap<K, V> {
     /// ```
     pub fn insert(&mut self, key: K, value: V) -> Option<V> {
         let key = key.into();
-        for _ in self.slots.len()..=key.idx as usize {
-            self.slots.push(Slot {
+        self.slots
+            .extend((self.slots.len()..=key.idx as usize).map(|_| Slot {
                 version: 0,
                 value: None,
-            });
-        }
+            }));
 
         let slot = &mut self.slots[key.idx as usize];
         if slot.version == key.version.get() {
@@ -1020,7 +1049,7 @@ mod serialize {
             D: Deserializer<'de>,
         {
             let mut slots: Vec<Slot<V>> = Deserialize::deserialize(deserializer)?;
-            if slots.len() >= (1 << 32) - 1 {
+            if slots.len() >= u32::max_value() as usize {
                 return Err(de::Error::custom(&"too many slots"));
             }
 
@@ -1047,8 +1076,8 @@ mod serialize {
 
 #[cfg(test)]
 mod tests {
+    use crate::*;
     use std::collections::HashMap;
-    use *;
 
     #[cfg(feature = "serde")]
     use serde_json;
@@ -1063,9 +1092,9 @@ mod tests {
             let mut sm_keys = Vec::new();
 
             #[cfg(not(feature = "serde"))]
-            let num_ops = 3;
-            #[cfg(feature = "serde")]
             let num_ops = 4;
+            #[cfg(feature = "serde")]
+            let num_ops = 5;
 
             for (op, val) in operations {
                 match op % num_ops {
@@ -1103,9 +1132,14 @@ mod tests {
                         }
                     }
 
+                    // Clone.
+                    3 => {
+                        sec = sec.clone();
+                    }
+
                     // Serde round-trip.
                     #[cfg(feature = "serde")]
-                    3 => {
+                    4 => {
                         let ser = serde_json::to_string(&sec).unwrap();
                         sec = serde_json::from_str(&ser).unwrap();
                     }
