@@ -1,14 +1,16 @@
 //! Contains the sparse secondary map implementation.
 
 use super::{is_older_version, Key, KeyData};
-use std;
-use std::collections::hash_map;
-use std::collections::hash_map::HashMap;
+use std::collections::hash_map::{self, HashMap};
+use std::hash;
 use std::iter::{Extend, FromIterator, FusedIterator};
 use std::marker::PhantomData;
 use std::ops::{Index, IndexMut};
 
-#[derive(Debug)]
+#[cfg(feature = "unstable")]
+use std::collections::TryReserveError;
+
+#[derive(Debug, Clone)]
 struct Slot<T> {
     version: u32,
     value: T,
@@ -71,13 +73,13 @@ struct Slot<T> {
 /// ammo[alice] = 0;
 /// ```
 
-#[derive(Debug)]
-pub struct SparseSecondaryMap<K: Key, V> {
-    slots: HashMap<u32, Slot<V>>,
+#[derive(Debug, Clone)]
+pub struct SparseSecondaryMap<K: Key, V, S: hash::BuildHasher = hash_map::RandomState> {
+    slots: HashMap<u32, Slot<V>, S>,
     _k: PhantomData<fn(K) -> K>,
 }
 
-impl<K: Key, V> SparseSecondaryMap<K, V> {
+impl<K: Key, V> SparseSecondaryMap<K, V, hash_map::RandomState> {
     /// Constructs a new, empty `SparseSecondaryMap`.
     ///
     /// # Examples
@@ -109,6 +111,52 @@ impl<K: Key, V> SparseSecondaryMap<K, V> {
             _k: PhantomData,
         }
     }
+}
+
+impl<K: Key, V, S: hash::BuildHasher> SparseSecondaryMap<K, V, S> {
+    /// Creates an empty `SparseSecondaryMap` which will use the given hash
+    /// builder to hash keys.
+    ///
+    /// The secondary map will not reallocate until it holds at least `capacity`
+    /// slots.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// # use std::collections::hash_map::RandomState;
+    /// # use slotmap::*;
+    /// let mut sm: SlotMap<_, i32> = SlotMap::with_capacity(10);
+    /// let mut sec: SparseSecondaryMap<DefaultKey, i32, _> =
+    ///     SparseSecondaryMap::with_hasher(RandomState::new());
+    /// ```
+    pub fn with_hasher(hash_builder: S) -> Self {
+        Self {
+            slots: HashMap::with_hasher(hash_builder),
+            _k: PhantomData,
+        }
+    }
+
+    /// Creates an empty `SparseSecondaryMap` with the given capacity of slots,
+    /// using `hash_builder` to hash the keys.
+    ///
+    /// The secondary map will not reallocate until it holds at least `capacity`
+    /// slots.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// # use std::collections::hash_map::RandomState;
+    /// # use slotmap::*;
+    /// let mut sm: SlotMap<_, i32> = SlotMap::with_capacity(10);
+    /// let mut sec: SparseSecondaryMap<DefaultKey, i32, _> =
+    ///     SparseSecondaryMap::with_capacity_and_hasher(10, RandomState::new());
+    /// ```
+    pub fn with_capacity_and_hasher(capacity: usize, hash_builder: S) -> Self {
+        Self {
+            slots: HashMap::with_capacity_and_hasher(capacity, hash_builder),
+            _k: PhantomData,
+        }
+    }
 
     /// Returns the number of elements in the secondary map.
     ///
@@ -166,15 +214,31 @@ impl<K: Key, V> SparseSecondaryMap<K, V> {
     ///
     /// ```
     /// # use slotmap::*;
-    /// let mut sec: SparseSecondaryMap<DefaultKey, i32> = SparseSecondaryMap::with_capacity(10);
+    /// let mut sec: SparseSecondaryMap<DefaultKey, i32> = SparseSecondaryMap::new();
+    /// sec.reserve(10);
     /// assert!(sec.capacity() >= 10);
-    /// sec.reserve(20);
-    /// assert!(sec.capacity() >= 20);
     /// ```
     pub fn reserve(&mut self, additional: usize) {
         self.slots.reserve(additional);
     }
 
+    /// Tries to reserve capacity for at least `additional` more slots in the
+    /// `SparseSecondaryMap`.  The collection may reserve more space to avoid
+    /// frequent reallocations.
+    ///
+    /// # Examples
+    ///
+    /// ```
+    /// # use slotmap::*;
+    /// let mut sec: SparseSecondaryMap<DefaultKey, i32> = SparseSecondaryMap::new();
+    /// sec.try_reserve(10).unwrap();
+    /// assert!(sec.capacity() >= 10);
+    /// ```
+    #[cfg(feature = "unstable")]
+    pub fn try_reserve(&mut self, additional: usize) -> Result<(), TryReserveError> {
+        self.slots.try_reserve(additional)
+    }
+
     /// Returns `true` if the secondary map contains `key`.
     ///
     /// # Examples
@@ -542,13 +606,21 @@ impl<K: Key, V> SparseSecondaryMap<K, V> {
     }
 }
 
-impl<K: Key, V> Default for SparseSecondaryMap<K, V> {
+impl<K, V, S> Default for SparseSecondaryMap<K, V, S>
+where
+    K: Key,
+    S: hash::BuildHasher + Default,
+{
     fn default() -> Self {
-        Self::new()
+        Self::with_hasher(Default::default())
     }
 }
 
-impl<K: Key, V> Index<K> for SparseSecondaryMap<K, V> {
+impl<K, V, S> Index<K> for SparseSecondaryMap<K, V, S>
+where
+    K: Key,
+    S: hash::BuildHasher,
+{
     type Output = V;
 
     fn index(&self, key: K) -> &V {
@@ -559,7 +631,11 @@ impl<K: Key, V> Index<K> for SparseSecondaryMap<K, V> {
     }
 }
 
-impl<K: Key, V> IndexMut<K> for SparseSecondaryMap<K, V> {
+impl<K, V, S> IndexMut<K> for SparseSecondaryMap<K, V, S>
+where
+    K: Key,
+    S: hash::BuildHasher,
+{
     fn index_mut(&mut self, key: K) -> &mut V {
         match self.get_mut(key) {
             Some(r) => r,
@@ -568,7 +644,12 @@ impl<K: Key, V> IndexMut<K> for SparseSecondaryMap<K, V> {
     }
 }
 
-impl<K: Key, V: PartialEq> PartialEq for SparseSecondaryMap<K, V> {
+impl<K, V, S> PartialEq for SparseSecondaryMap<K, V, S>
+where
+    K: Key,
+    V: PartialEq,
+    S: hash::BuildHasher,
+{
     fn eq(&self, other: &Self) -> bool {
         if self.len() != other.len() {
             return false;
@@ -582,17 +663,31 @@ impl<K: Key, V: PartialEq> PartialEq for SparseSecondaryMap<K, V> {
     }
 }
 
-impl<K: Key, V: Eq> Eq for SparseSecondaryMap<K, V> {}
+impl<K, V, S> Eq for SparseSecondaryMap<K, V, S>
+where
+    K: Key,
+    V: Eq,
+    S: hash::BuildHasher,
+{
+}
 
-impl<K: Key, V> FromIterator<(K, V)> for SparseSecondaryMap<K, V> {
+impl<K, V, S> FromIterator<(K, V)> for SparseSecondaryMap<K, V, S>
+where
+    K: Key,
+    S: hash::BuildHasher + Default,
+{
     fn from_iter<I: IntoIterator<Item = (K, V)>>(iter: I) -> Self {
-        let mut sec = Self::new();
+        let mut sec = Self::default();
         sec.extend(iter);
         sec
     }
 }
 
-impl<K: Key, V> Extend<(K, V)> for SparseSecondaryMap<K, V> {
+impl<K, V, S> Extend<(K, V)> for SparseSecondaryMap<K, V, S>
+where
+    K: Key,
+    S: hash::BuildHasher,
+{
     fn extend<I: IntoIterator<Item = (K, V)>>(&mut self, iter: I) {
         let iter = iter.into_iter();
         for (k, v) in iter {
@@ -601,7 +696,12 @@ impl<K: Key, V> Extend<(K, V)> for SparseSecondaryMap<K, V> {
     }
 }
 
-impl<'a, K: Key, V: 'a + Copy> Extend<(K, &'a V)> for SparseSecondaryMap<K, V> {
+impl<'a, K, V, S> Extend<(K, &'a V)> for SparseSecondaryMap<K, V, S>
+where
+    K: Key,
+    V: 'a + Copy,
+    S: hash::BuildHasher,
+{
     fn extend<I: IntoIterator<Item = (K, &'a V)>>(&mut self, iter: I) {
         let iter = iter.into_iter();
         for (k, v) in iter {
@@ -759,7 +859,11 @@ impl<'a, K: Key, V> Iterator for ValuesMut<'a, K, V> {
     }
 }
 
-impl<'a, K: Key, V> IntoIterator for &'a SparseSecondaryMap<K, V> {
+impl<'a, K, V, S> IntoIterator for &'a SparseSecondaryMap<K, V, S>
+where
+    K: Key,
+    S: hash::BuildHasher,
+{
     type Item = (K, &'a V);
     type IntoIter = Iter<'a, K, V>;
 
@@ -768,7 +872,11 @@ impl<'a, K: Key, V> IntoIterator for &'a SparseSecondaryMap<K, V> {
     }
 }
 
-impl<'a, K: Key, V> IntoIterator for &'a mut SparseSecondaryMap<K, V> {
+impl<'a, K, V, S> IntoIterator for &'a mut SparseSecondaryMap<K, V, S>
+where
+    K: Key,
+    S: hash::BuildHasher,
+{
     type Item = (K, &'a mut V);
     type IntoIter = IterMut<'a, K, V>;
 
@@ -777,7 +885,11 @@ impl<'a, K: Key, V> IntoIterator for &'a mut SparseSecondaryMap<K, V> {
     }
 }
 
-impl<K: Key, V> IntoIterator for SparseSecondaryMap<K, V> {
+impl<K, V, S> IntoIterator for SparseSecondaryMap<K, V, S>
+where
+    K: Key,
+    S: hash::BuildHasher,
+{
     type Item = (K, V);
     type IntoIter = IntoIter<K, V>;
 
@@ -809,10 +921,15 @@ impl<K: Key, V> ExactSizeIterator for IntoIter<K, V> {}
 #[cfg(feature = "serde")]
 mod serialize {
     use super::*;
+    use crate::SecondaryMap;
     use serde::{Deserialize, Deserializer, Serialize, Serializer};
-    use SecondaryMap;
 
-    impl<K: Key, V: Serialize> Serialize for SparseSecondaryMap<K, V> {
+    impl<K, V, H> Serialize for SparseSecondaryMap<K, V, H>
+    where
+        K: Key,
+        V: Serialize,
+        H: hash::BuildHasher,
+    {
         fn serialize<S>(&self, serializer: S) -> Result<S::Ok, S::Error>
         where
             S: Serializer,
@@ -826,13 +943,18 @@ mod serialize {
         }
     }
 
-    impl<'de, K: Key, V: Deserialize<'de>> Deserialize<'de> for SparseSecondaryMap<K, V> {
+    impl<'de, K, V, S> Deserialize<'de> for SparseSecondaryMap<K, V, S>
+    where
+        K: Key,
+        V: Deserialize<'de>,
+        S: hash::BuildHasher + Default,
+    {
         fn deserialize<D>(deserializer: D) -> Result<Self, D::Error>
         where
             D: Deserializer<'de>,
         {
             let serde_sec: SecondaryMap<K, V> = Deserialize::deserialize(deserializer)?;
-            let mut sec = Self::new();
+            let mut sec = Self::default();
 
             for (k, v) in serde_sec {
                 sec.insert(k, v);
@@ -845,12 +967,28 @@ mod serialize {
 
 #[cfg(test)]
 mod tests {
+    use crate::*;
     use std::collections::HashMap;
-    use *;
 
     #[cfg(feature = "serde")]
     use serde_json;
 
+    #[test]
+    fn custom_hasher() {
+        type FastSparseSecondaryMap<K, V> = SparseSecondaryMap<K, V, fxhash::FxBuildHasher>;
+        let mut sm = SlotMap::new();
+        let mut sec = FastSparseSecondaryMap::default();
+        let key1 = sm.insert(42);
+        sec.insert(key1, 1234);
+        assert_eq!(sec[key1], 1234);
+        assert_eq!(sec.len(), 1);
+        let sec2 = sec
+            .iter()
+            .map(|(k, &v)| (k, v))
+            .collect::<FastSparseSecondaryMap<_, _>>();
+        assert_eq!(sec, sec2);
+    }
+
     quickcheck! {
         fn qc_secmap_equiv_hashmap(operations: Vec<(u8, u32)>) -> bool {
             let mut hm = HashMap::new();
@@ -861,9 +999,9 @@ mod tests {
             let mut sm_keys = Vec::new();
 
             #[cfg(not(feature = "serde"))]
-            let num_ops = 3;
-            #[cfg(feature = "serde")]
             let num_ops = 4;
+            #[cfg(feature = "serde")]
+            let num_ops = 5;
 
             for (op, val) in operations {
                 match op % num_ops {
@@ -901,9 +1039,14 @@ mod tests {
                         }
                     }
 
+                    // Clone.
+                    3 => {
+                        sec = sec.clone();
+                    }
+
                     // Serde round-trip.
                     #[cfg(feature = "serde")]
-                    3 => {
+                    4 => {
                         let ser = serde_json::to_string(&sec).unwrap();
                         sec = serde_json::from_str(&ser).unwrap();
                     }
